__author__ = "Felix Kleinert"
__date__ = "2021-01-29"

import os
import sys
from abc import ABC

import wget

from mlair.configuration import check_path_and_create


class AbstractReferenceModel(ABC):
    """
    Abstract reference model. All classes providing some reference or competitor models must inherent from this class.
    """
    def __init__(self, *args, **kwargs):
        pass

    def make_reference_available_locally(self):
        raise NotImplementedError

    @staticmethod
    def is_reference_available_locally(reference_path) -> bool:
        """
        Checks if reference is available locally
        """

        try:
            if os.listdir(reference_path):
                res = True
            else:
                res = False
        except FileNotFoundError:
            res = False
        return res


class AbstractReferenceB2share(AbstractReferenceModel):
    """
    Abstract class for reference models located on b2share (eudat or fz-juelich)
    See also https://github.com/EUDAT-Training/B2SHARE-Training/blob/master/api/01_Retrieve_existing_record.md

    """

    def __init__(self, b2share_hosturl: str, b2share_bucket: str, b2share_key: str):
        super().__init__()
        self.b2share_hosturl = b2share_hosturl
        self.b2share_bucket = b2share_bucket
        self.b2share_key = b2share_key

    @property
    def b2share_url(self):
        return f"{self.b2share_hosturl}/api/files/{self.b2share_bucket}"

    def bar_custom(self, current, total, width=80):
        progress_message = f"Downloading {self.b2share_key}: {round(current / total * 100)}% [{current} / {total}] bytes"
        sys.stdout.write("\r" + progress_message)
        sys.stdout.flush()

    def download_from_b2share(self, tmp_download_path: str):
        check_path_and_create(tmp_download_path)
        wget.download(f"{self.b2share_url}/{self.b2share_key}",
                      out=f"{tmp_download_path}{self.b2share_key}",
                      bar=self.bar_custom
                      )

    def make_reference_available_locally(self):
        raise NotImplementedError
