import keras
import numpy as np

from mlair.model_modules.loss import l_p_loss, var_loss, custom_loss

import pytest


class TestLPLoss:

    def test_l_p_loss(self):
        model = keras.Sequential()
        model.add(keras.layers.Lambda(lambda x: x, input_shape=(None,)))
        model.compile(optimizer=keras.optimizers.Adam(), loss=l_p_loss(2))
        hist = model.fit(np.array([1, 0, 2, 0.5]), np.array([1, 1, 0, 0.5]), epochs=1)
        assert hist.history['loss'][0] == 1.25
        model.compile(optimizer=keras.optimizers.Adam(), loss=l_p_loss(3))
        hist = model.fit(np.array([1, 0, -2, 0.5]), np.array([1, 1, 0, 0.5]), epochs=1)
        assert hist.history['loss'][0] == 2.25


class TestVarLoss:

    def test_var_loss(self):
        model = keras.Sequential()
        model.add(keras.layers.Lambda(lambda x: x, input_shape=(None,)))
        model.compile(optimizer=keras.optimizers.Adam(), loss=var_loss)
        hist = model.fit(np.array([1, 0, 2, 0.5]), np.array([1, 1, 0, 0.5]), epochs=1)
        assert hist.history['loss'][0] == 0.140625


class TestCustomLoss:

    def test_custom_loss_no_weights(self):
        cust_loss = custom_loss([l_p_loss(2), var_loss])
        model = keras.Sequential()
        model.add(keras.layers.Lambda(lambda x: x, input_shape=(None,)))
        model.compile(optimizer=keras.optimizers.Adam(), loss=cust_loss)
        hist = model.fit(np.array([1, 0, 2, 0.5]), np.array([1, 1, 0, 0.5]), epochs=1)
        assert hist.history['loss'][0] == (0.5 * 0.140625 + 0.5 * 1.25)

    @pytest.mark.parametrize("weights", [[0.3, 0.7], [0.5, 0.5], [1, 1], [4, 1]])
    def test_custom_loss_with_weights(self, weights):
        cust_loss = custom_loss([l_p_loss(2), var_loss], weights)
        model = keras.Sequential()
        model.add(keras.layers.Lambda(lambda x: x, input_shape=(None,)))
        model.compile(optimizer=keras.optimizers.Adam(), loss=cust_loss)
        hist = model.fit(np.array([1, 0, 2, 0.5]), np.array([1, 1, 0, 0.5]), epochs=1)
        weights_adjusted = list(map(lambda x: x / sum(weights), weights))
        expected = (weights_adjusted[0] * 1.25 + weights_adjusted[1] * 0.140625)
        assert np.testing.assert_almost_equal(hist.history['loss'][0], expected, decimal=6) is None

    def test_custom_loss_invalid_weights(self):
        with pytest.raises(AssertionError):
            custom_loss([l_p_loss(2), var_loss], [0.3])
        with pytest.raises(AssertionError):
            custom_loss([l_p_loss(2), var_loss], [0.4, 3, 1])
